@@ -1,10 +1,10 @@
 /*******************************************************************************
  *
  * NRPE.C - Nagios Remote Plugin Executor
- * Copyright (c) 1999-2003 Ethan Galstad (nagios@nagios.org)
+ * Copyright (c) 1999-2006 Ethan Galstad (nagios@nagios.org)
  * License: GPL
  *
- * Last Modified: 07-22-2003
+ * Last Modified: 01-23-2006
  *
  * Command line: nrpe -c <config_file> [--inetd | --daemon]
  *
@@ -18,33 +18,36 @@
  * 
  ******************************************************************************/
 
-#include "../common/common.h"
-#include "../common/config.h"
-#include "nrpe.h"
-#include "utils.h"
+#include "../include/common.h"
+#include "../include/config.h"
+#include "../include/nrpe.h"
+#include "../include/utils.h"
+
+#ifdef HAVE_SSL
+#include "../include/dh.h"
+#endif
+
+#ifdef HAVE_LIBWRAP
+int allow_severity=LOG_INFO;
+int deny_severity=LOG_WARNING;
+#endif
+
+#ifdef HAVE_SSL
+SSL_METHOD *meth;
+SSL_CTX *ctx;
+int use_ssl=TRUE;
+#else
+int use_ssl=FALSE;
+#endif
 
 #define DEFAULT_COMMAND_TIMEOUT	60			/* default timeout for execution of plugins */
 #define MAXFD                   64
+#define NASTY_METACHARS         "|`&><'\"\\[]{}"
 
-
-int process_arguments(int,char **);
-
-void wait_for_connections(void);
-void handle_connection(int);
-int read_config_file(char *);
-int add_command(char *,char *);
-command *find_command(char *);
-void sighandler(int);
-int drop_privileges(char *,char *);
-void free_memory(void);
-int is_an_allowed_host(char *);
-
-int my_system(char *,int,int *,char *,int);            	/* executes a command via popen(), but also protects against timeouts */
-void my_system_sighandler(int);				/* handles timeouts when executing commands via my_system() */
-
+char    *command_name=NULL;
+char    *macro_argv[MAX_COMMAND_ARGUMENTS];
 
 char    config_file[MAX_INPUT_BUFFER]="nrpe.cfg";
-char    allowed_hosts[MAX_INPUT_BUFFER];
 int     server_port=DEFAULT_SERVER_PORT;
 char    server_address[16]="0.0.0.0";
 int     socket_timeout=DEFAULT_SOCKET_TIMEOUT;
@@ -55,6 +58,10 @@ command *command_list=NULL;
 char    *nrpe_user=NULL;
 char    *nrpe_group=NULL;
 
+char    *pid_file=NULL;
+
+int     allow_arguments=FALSE;
+
 int     show_help=FALSE;
 int     show_license=FALSE;
 int     show_version=FALSE;
@@ -63,11 +70,14 @@ int     debug=FALSE;
 
 
 
+
 int main(int argc, char **argv){
-	int error=FALSE;
-	int result;
-	int i;
+	int result=OK;
+	int x;
 	char buffer[MAX_INPUT_BUFFER];
+#ifdef HAVE_SSL
+	DH *dh;
+#endif
 
 	result=process_arguments(argc,argv);
 
@@ -75,22 +85,46 @@ int main(int argc, char **argv){
 
 		printf("\n");
 		printf("NRPE - Nagios Remote Plugin Executor\n");
-		printf("Copyright (c) 1999-2003 Ethan Galstad (nagios@nagios.org)\n");
+		printf("Copyright (c) 1999-2006 Ethan Galstad (nagios@nagios.org)\n");
 		printf("Version: %s\n",PROGRAM_VERSION);
 		printf("Last Modified: %s\n",MODIFICATION_DATE);
-		printf("License: GPL\n");
+		printf("License: GPL v2 with exemptions (-l for more info)\n");
+#ifdef HAVE_SSL
+		printf("SSL/TLS Available: Anonymous DH Mode, OpenSSL 0.9.6 or higher required\n");
+#endif
+#ifdef HAVE_LIBWRAP
+		printf("TCP Wrappers Available\n");
+#endif
 		printf("\n");
+#ifdef ENABLE_COMMAND_ARGUMENTS
+		printf("***************************************************************\n");
+		printf("** POSSIBLE SECURITY RISK - COMMAND ARGUMENTS ARE SUPPORTED! **\n");
+		printf("**      Read the NRPE SECURITY file for more information     **\n");
+		printf("***************************************************************\n");
+		printf("\n");
+#endif
+#ifndef HAVE_LIBWRAP
+		printf("***************************************************************\n");
+		printf("** POSSIBLE SECURITY RISK - TCP WRAPPERS ARE NOT AVAILABLE!  **\n");
+		printf("**      Read the NRPE SECURITY file for more information     **\n");
+		printf("***************************************************************\n");
+		printf("\n");
+#endif
 	        }
 
-	if(result!=OK || show_help==TRUE){
+	if(show_license==TRUE)
+		display_license();
 
-		printf("Usage: %s -c <config_file> [mode]\n",argv[0]);
+	else if(result!=OK || show_help==TRUE){
+
+		printf("Usage: nrpe [-n] -c <config_file> <mode>\n");
 		printf("\n");
 		printf("Options:\n");
+		printf(" -n            = Do not use SSL\n");
 		printf(" <config_file> = Name of config file to use\n");
-		printf(" [mode]        = Determines how NRPE should run. Valid modes:\n");
-		printf("   --inetd     = Run as a service under inetd or xinetd\n");
-		printf("   --daemon    = Run as a standalone daemon\n");
+		printf(" <mode>        = One of the following two operating modes:\n");  
+		printf("   -i          =    Run as a service under inetd or xinetd\n");
+		printf("   -d          =    Run as a standalone daemon\n");
 		printf("\n");
 		printf("Notes:\n");
 		printf("This program is designed to process requests from the check_nrpe\n");
@@ -103,9 +137,6 @@ int main(int argc, char **argv){
 		printf("\n");
 		}
 
-	if(show_license==TRUE)
-		display_license();
-
         if(result!=OK || show_help==TRUE || show_license==TRUE || show_version==TRUE)
 		exit(STATE_UNKNOWN);
 
@@ -142,21 +173,72 @@ int main(int argc, char **argv){
 		return STATE_CRITICAL;
 		}
 
+	/* initialize macros */
+	for(x=0;x<MAX_COMMAND_ARGUMENTS;x++)
+		macro_argv[x]=NULL;
+
+        /* generate the CRC 32 table */
+        generate_crc32_table();
+
+#ifdef HAVE_SSL
+	/* initialize SSL */
+	if(use_ssl==TRUE){
+		SSL_library_init();
+		SSLeay_add_ssl_algorithms();
+		meth=SSLv23_server_method();
+		SSL_load_error_strings();
+		if((ctx=SSL_CTX_new(meth))==NULL){
+			syslog(LOG_ERR,"Error: could not create SSL context.\n");
+			exit(STATE_CRITICAL);
+		        }
+
+		/* ADDED 01/19/2004 */
+		/* use only TLSv1 protocol */
+		SSL_CTX_set_options(ctx,SSL_OP_NO_SSLv2 | SSL_OP_NO_SSLv3);
+
+		/* use anonymous DH ciphers */
+		SSL_CTX_set_cipher_list(ctx,"ADH");
+		dh=get_dh512();
+		SSL_CTX_set_tmp_dh(ctx,dh);
+		DH_free(dh);
+		if(debug==TRUE)
+			syslog(LOG_INFO,"INFO: SSL/TLS initialized. All network traffic will be encrypted.");
+	        }
+	else{
+		if(debug==TRUE)
+			syslog(LOG_INFO,"INFO: SSL/TLS NOT initialized. Network encryption DISABLED.");
+	        }
+#endif
+
 	/* if we're running under inetd... */
-	if(use_inetd==TRUE)
+	if(use_inetd==TRUE){
+
+		/* make sure we're not root */
+		check_privileges();
+
+		/* redirect STDERR to /dev/null */
+		close(2);
+		open("/dev/null",O_WRONLY);
+
+		/* handle the connection */
 		handle_connection(0);
+	        }
 
 	/* else daemonize and start listening for requests... */
 	else if(fork()==0){
-
+		
 		/* we're a daemon - set up a new process group */
 		setsid();
 
 		/* ignore SIGHUP */
 		signal(SIGHUP, SIG_IGN);
 
+		/* write pid file */
+		if(write_pid_file()==ERROR)
+			return STATE_CRITICAL;
+
 		chdir("/");
-		umask(0);
+		/*umask(0);*/
 
 		/* close standard file descriptors */
 		close(0);
@@ -171,10 +253,18 @@ int main(int argc, char **argv){
 		/* drop privileges */
 		drop_privileges(nrpe_user,nrpe_group);
 
+		/* make sure we're not root */
+		check_privileges();
+
 		/* wait for connections */
 		wait_for_connections();
 	        }
 
+#ifdef HAVE_SSL
+	if(use_ssl==TRUE)
+		SSL_CTX_free(ctx);
+#endif
+
 	/* We are now running in daemon mode, or the connection handed over by inetd has
 	   been completed, so the parent process exits */
         return STATE_OK;
@@ -186,11 +276,15 @@ int main(int argc, char **argv){
 /* read in the configuration file */
 int read_config_file(char *filename){
 	FILE *fp;
+	char config_file[MAX_FILENAME_LENGTH];
 	char input_buffer[MAX_INPUT_BUFFER];
+	char *input_line;
 	char *temp_buffer;
 	char *varname;
 	char *varvalue;
-	int line;
+	int line=0;
+	int len=0;
+	int x=0;
 
 
 	/* open the config file for reading */
@@ -202,21 +296,34 @@ int read_config_file(char *filename){
 		return ERROR;
 	        }
 
-	line=0;
 	while(fgets(input_buffer,MAX_INPUT_BUFFER-1,fp)){
 
 		line++;
+		input_line=input_buffer;
+
+		/* skip leading whitespace */
+		while(isspace(*input_line))
+			++input_line;
+
+		/* trim trailing whitespace */
+		len=strlen(input_line);
+		for(x=len-1;x>=0;x--){
+			if(isspace(input_line[x]))
+				input_line[x]='\x0';
+			else
+				break;
+		        }
 
 		/* skip comments and blank lines */
-		if(input_buffer[0]=='#')
+		if(input_line[0]=='#')
 			continue;
-		if(input_buffer[0]=='\x0')
+		if(input_line[0]=='\x0')
 			continue;
-		if(input_buffer[0]=='\n')
+		if(input_line[0]=='\n')
 			continue;
 
 		/* get the variable name */
-		varname=strtok(input_buffer,"=");
+		varname=strtok(input_line,"=");
 		if(varname==NULL){
 			syslog(LOG_ERR,"No variable name specified in config file '%s' - Line %d\n",filename,line);
 			return ERROR;
@@ -229,7 +336,30 @@ int read_config_file(char *filename){
 			return ERROR;
 		        }
 
-		if(!strcmp(varname,"server_port")){
+		/* allow users to specify directories to recurse into for config files */
+		else if(!strcmp(varname,"include_dir")){
+
+			strncpy(config_file,varvalue,sizeof(config_file)-1);
+			config_file[sizeof(config_file)-1]='\x0';
+
+			/* strip trailing / if necessary */
+			if(config_file[strlen(config_file)-1]=='/')
+				config_file[strlen(config_file)-1]='\x0';
+
+			/* process the config directory... */
+			if(read_config_dir(config_file)==ERROR)
+				break;
+		        }
+
+		/* allow users to specify individual config files to include */
+		else if(!strcmp(varname,"include") || !strcmp(varname,"include_file")){
+
+			/* process the config file... */
+			if(read_config_file(varvalue)==ERROR)
+				break;
+		        }
+
+		else if(!strcmp(varname,"server_port")){
 			server_port=atoi(varvalue);
 			if(server_port<1024){
 				syslog(LOG_ERR,"Invalid port number specified in config file '%s' - Line %d\n",filename,line);
@@ -237,21 +367,12 @@ int read_config_file(char *filename){
 			        }
 		        }
 
-                else if(!strcmp(varname,"server_address")){
+		else if(!strcmp(varname,"server_address")){
                         strncpy(server_address,varvalue,sizeof(server_address) - 1);
-                        server_address[sizeof(server_address) - 1] = '\0';
+                        server_address[sizeof(server_address)-1]='\0';
                         }
 
-		else if(!strcmp(varname,"allowed_hosts")){
-			if(strlen(input_buffer)>sizeof(allowed_hosts)){
-				syslog(LOG_ERR,"Allowed hosts list too long in config file '%s' - Line %d\n",filename,line);
-				return ERROR;
-			        }
-			strncpy(allowed_hosts,varvalue,sizeof(allowed_hosts));
-			allowed_hosts[sizeof(allowed_hosts)-1]='\x0';
-		        }
-
-		else if(strstr(input_buffer,"command[")){
+		else if(strstr(input_line,"command[")){
 			temp_buffer=strtok(varname,"[");
 			temp_buffer=strtok(NULL,"]");
 			if(temp_buffer==NULL){
@@ -275,7 +396,10 @@ int read_config_file(char *filename){
                 else if(!strcmp(varname,"nrpe_group"))
 			nrpe_group=strdup(varvalue);
 		
-		else if(!strcmp(varname,"command_timeout")){
+		else if(!strcmp(varname,"dont_blame_nrpe"))
+			allow_arguments=(atoi(varvalue)==1)?TRUE:FALSE;
+
+ 		else if(!strcmp(varname,"command_timeout")){
 			command_timeout=atoi(varvalue);
 			if(command_timeout<1){
 				syslog(LOG_ERR,"Invalid command_timeout specified in config file '%s' - Line %d\n",filename,line);
@@ -283,10 +407,12 @@ int read_config_file(char *filename){
 			        }
 		        }
 
-		else{
-			syslog(LOG_ERR,"Unknown option specified in config file '%s' - Line %d\n",filename,line);
+		else if(!strcmp(varname,"pid_file"))
+			pid_file=strdup(varvalue);
 
-			return ERROR;
+		else{
+			syslog(LOG_WARNING,"Unknown option specified in config file '%s' - Line %d\n",filename,line);
+			continue;
 		        }
 
 	        }
@@ -299,18 +425,97 @@ int read_config_file(char *filename){
 	}
 
 
+/* process all config files in a specific config directory (with directory recursion) */
+int read_config_dir(char *dirname){
+	char config_file[MAX_FILENAME_LENGTH];
+	DIR *dirp;
+	struct dirent *dirfile;
+	int result=OK;
+	int x;
+
+	/* open the directory for reading */
+	dirp=opendir(dirname);
+        if(dirp==NULL){
+		syslog(LOG_ERR,"Could not open config directory '%s' for reading.\n",dirname);
+		return ERROR;
+	        }
+
+	/* process all files in the directory... */
+	while((dirfile=readdir(dirp))!=NULL){
+
+		/* process this if it's a config file... */
+		x=strlen(dirfile->d_name);
+		if(x>4 && !strcmp(dirfile->d_name+(x-4),".cfg")){
+
+#ifdef _DIRENT_HAVE_D_TYPE
+			/* only process normal files */
+			if(dirfile->d_type!=DT_REG)
+				continue;
+#endif
+
+			/* create the full path to the config file */
+			snprintf(config_file,sizeof(config_file)-1,"%s/%s",dirname,dirfile->d_name);
+			config_file[sizeof(config_file)-1]='\x0';
+
+			/* process the config file */
+			result=read_config_file(config_file);
+
+			/* break out if we encountered an error */
+			if(result==ERROR)
+				break;
+		        }
+
+#ifdef _DIRENT_HAVE_D_TYPE
+		/* recurse into subdirectories... */
+		if(dirfile->d_type==DT_DIR){
+
+			/* ignore current, parent and hidden directory entries */
+			if(dirfile->d_name[0]=='.')
+				continue;
+
+			/* create the full path to the config directory */
+			snprintf(config_file,sizeof(config_file)-1,"%s/%s",dirname,dirfile->d_name);
+			config_file[sizeof(config_file)-1]='\x0';
+
+			/* process the config directory */
+			result=read_config_dir(config_file);
+
+			/* break out if we encountered an error */
+			if(result==ERROR)
+				break;
+		        }
+#endif
+		}
+
+	closedir(dirp);
+
+	return result;
+        }
+
 
 /* adds a new command definition from the config file to the list in memory */
 int add_command(char *command_name, char *command_line){
 	command *new_command;
 
+	if(command_name==NULL || command_line==NULL)
+		return ERROR;
+
 	/* allocate memory for the new command */
 	new_command=(command *)malloc(sizeof(command));
 	if(new_command==NULL)
 		return ERROR;
 
-	strcpy(new_command->command_name,command_name);
-	strcpy(new_command->command_line,command_line);
+	new_command->command_name=strdup(command_name);
+	if(new_command->command_name==NULL){
+		free(new_command);
+		return ERROR;
+	        }
+	new_command->command_line=strdup(command_line);
+	if(new_command->command_line==NULL){
+		free(new_command->command_name);
+		free(new_command);
+		return ERROR;
+	        }
 
 	/* add new command to head of list in memory */
 	new_command->next=command_list;
@@ -343,10 +548,17 @@ void wait_for_connections(void){
 	struct sockaddr_in *nptr;
 	struct sockaddr addr;
 	int rc;
-	int sock, new_sd, addrlen;
+	int sock, new_sd;
+	socklen_t addrlen;
 	char connecting_host[16];
 	pid_t pid;
 	int flag=1;
+	fd_set fdread;
+	struct timeval timeout;
+	int retval;
+#ifdef HAVE_LIBWRAP
+	struct request_info req;
+#endif
 
 	/* create a socket for listening */
 	sock=socket(AF_INET,SOCK_STREAM,0);
@@ -354,7 +566,7 @@ void wait_for_connections(void){
 	/* exit if we couldn't create the socket */
 	if(sock<0){
 	        syslog(LOG_ERR,"Network server socket failure (%d: %s)",errno,strerror(errno));
-	        exit (STATE_CRITICAL);
+	        exit(STATE_CRITICAL);
 		}
 
         /* set the reuse address flag so we don't get errors when restarting */
@@ -374,31 +586,36 @@ void wait_for_connections(void){
 
 	else if(!my_inet_aton(server_address,&myname.sin_addr)){
 		syslog(LOG_ERR,"Server address is not a valid IP address\n");
-		exit (STATE_CRITICAL);
+		exit(STATE_CRITICAL);
                 }
 
 	/* bind the address to the Internet socket */
 	if(bind(sock,(struct sockaddr *)&myname,sizeof(myname))<0){
 		syslog(LOG_ERR,"Network server bind failure (%d: %s)\n",errno,strerror(errno));
-	        exit (STATE_CRITICAL);
+	        exit(STATE_CRITICAL);
 	        }
 
 	/* open the socket for listening */
 	if(listen(sock,5)<0){
 	    	syslog(LOG_ERR,"Network server listen failure (%d: %s)\n",errno,strerror(errno));
-	        exit (STATE_CRITICAL);
+	        exit(STATE_CRITICAL);
 		}
 
 	/* log info to syslog facility */
         syslog(LOG_NOTICE,"Starting up daemon");
 
+	/* log warning about command arguments */
+#ifdef ENABLE_COMMAND_ARGUMENTS
+	if(allow_arguments==TRUE)
+		syslog(LOG_NOTICE,"Warning: Daemon is configured to accept command arguments from clients!");
+#endif
+
 	/* Trap signals */
 	signal(SIGQUIT,sighandler);
 	signal(SIGTERM,sighandler);
 
 	if(debug==TRUE){
 		syslog(LOG_DEBUG,"Listening for connections on port %d\n",htons(myname.sin_port));
-		syslog(LOG_DEBUG,"Allowing connections from: %s\n",allowed_hosts);
 	        }
 
 	/* listen for connection requests - fork() if we get one */
@@ -406,10 +623,38 @@ void wait_for_connections(void){
 
 		/* wait for a connection request */
 	        while(1){
+
+			/* wait until there's something to do */
+			FD_ZERO(&fdread);
+			FD_SET(sock,&fdread);
+			timeout.tv_sec=0;
+			timeout.tv_usec=500000;
+			retval=select(sock+1,&fdread,NULL,&fdread,&timeout);
+
+			/* error */
+			if(retval<0)
+				continue;
+
+			/* accept a new connection request */
 			new_sd=accept(sock,0,0);
-			if(new_sd>=0 || (errno!=EWOULDBLOCK && errno!=EINTR))
+
+			/* some kind of error occurred... */
+			if(new_sd<0){
+
+				/* fix for HP-UX 11.0 - just retry */
+				if(errno==ENOBUFS)
+					continue;
+
+				/* retry */
+				if(errno==EWOULDBLOCK || errno==EINTR)
+					continue;
+
+				/* else handle the error later */
 				break;
-			sleep(1);
+			        }
+
+			/* connection was good */
+			break;
 		        }
 
 		/* child process should handle the connection */
@@ -426,9 +671,9 @@ void wait_for_connections(void){
 					/* log error to syslog facility */
 					syslog(LOG_ERR,"Network server accept failure (%d: %s)",errno,strerror(errno));
 
-					/* close socket prior to exiting */
+					/* close socket prioer to exiting */
 					close(sock);
-
+			
 					return;
 				        }
 
@@ -456,24 +701,28 @@ void wait_for_connections(void){
 				if(debug==TRUE)
 					syslog(LOG_DEBUG,"Connection from %s port %d",inet_ntoa(nptr->sin_addr),nptr->sin_port);
 
-				/* is this is a blessed machine? */
-				snprintf(connecting_host,sizeof(connecting_host),"%s",inet_ntoa(nptr->sin_addr));
-				connecting_host[sizeof(connecting_host)-1]='\x0';
+#ifdef HAVE_LIBWRAP
 
-				if(!is_an_allowed_host(connecting_host)){
+				/* Check whether or not connections are allowed from this host */
+				request_init(&req,RQ_DAEMON,"nrpe",RQ_FILE,new_sd,0);
+				fromhost(&req);
 
-				        /* log error to syslog facility */
-					syslog(LOG_ERR,"Host %s is not allowed to talk to us!",connecting_host);
-			                }
-				else{
+				if(!hosts_access(&req)){
 
-				        /* log info to syslog facility */
-					if(debug==TRUE)
-						syslog(LOG_DEBUG,"Host address checks out ok");
+					syslog(LOG_DEBUG,"Connection refused by TCP wrapper");
 
-				        /* handle the client connection */
-					handle_connection(new_sd);
-			                }
+					/* refuse the connection */
+					refuse(&req);
+					close(new_sd);
+
+					/* should not be reached */
+					syslog(LOG_ERR,"libwrap refuse() returns!");
+					exit(STATE_CRITICAL);
+					}
+#endif
+
+				/* handle the client connection */
+				handle_connection(new_sd);
 
 				/* log info to syslog facility */
 				if(debug==TRUE)
@@ -500,9 +749,6 @@ void wait_for_connections(void){
 		        }
   		}
 
-	/* we shouldn't ever get here... */
-	syslog(LOG_NOTICE,"Terminating");
-
 	return;
 	}
 
@@ -510,26 +756,70 @@ void wait_for_connections(void){
 
 /* handles a client connection */
 void handle_connection(int sock){
+        u_int32_t calculated_crc32;
 	command *temp_command;
 	packet receive_packet;
 	packet send_packet;
 	int bytes_to_send;
 	int bytes_to_recv;
 	char buffer[MAX_INPUT_BUFFER];
+	char raw_command[MAX_INPUT_BUFFER];
+	char processed_command[MAX_INPUT_BUFFER];
 	int result=STATE_OK;
 	int early_timeout=FALSE;
 	int rc;
+	int x;
+#ifdef DEBUG
+	FILE *errfp;
+#endif
+#ifdef HAVE_SSL
+	SSL *ssl=NULL;
+#endif
 
 
 	/* log info to syslog facility */
 	if(debug==TRUE)
 		syslog(LOG_DEBUG,"Handling the connection...");
 
+#ifdef OLDSTUFF
 	/* socket should be non-blocking */
 	fcntl(sock,F_SETFL,O_NONBLOCK);
+#endif
+
+#ifdef HAVE_SSL
+	/* do SSL handshake */
+	if(result==STATE_OK && use_ssl==TRUE){
+		if((ssl=SSL_new(ctx))!=NULL){
+			SSL_set_fd(ssl,sock);
+			if((rc=SSL_accept(ssl))!=1){
+				syslog(LOG_ERR,"Error: Could not complete SSL handshake. %d\n",SSL_get_error(ssl,rc));
+#ifdef DEBUG
+				errfp=fopen("/tmp/err.log","w");
+				ERR_print_errors_fp(errfp);
+				fclose(errfp);
+#endif
+				return;
+			        }
+		        }
+		else{
+			syslog(LOG_ERR,"Error: Could not create SSL connection structure.\n");
+#ifdef DEBUG
+			errfp=fopen("/tmp/err.log","w");
+			ERR_print_errors_fp(errfp);
+			fclose(errfp);
+#endif
+			return;
+		        }
+	        }
+#endif
 
 	bytes_to_recv=sizeof(receive_packet);
-	rc=recvall(sock,(char *)&receive_packet,&bytes_to_recv,socket_timeout);
+	if(use_ssl==FALSE)
+		rc=recvall(sock,(char *)&receive_packet,&bytes_to_recv,socket_timeout);
+#ifdef HAVE_SSL
+	else
+		rc=SSL_read(ssl,&receive_packet,bytes_to_recv);
+#endif
 
 	/* recv() error or client disconnect */
 	if(rc<=0){
@@ -537,6 +827,12 @@ void handle_connection(int sock){
 		/* log error to syslog facility */
 		syslog(LOG_ERR,"Could not read request from client, bailing out...");
 
+#ifdef HAVE_SSL
+		SSL_shutdown(ssl);
+		SSL_free(ssl);
+		syslog(LOG_INFO,"INFO: SSL Socket Shutdown.\n");
+#endif
+
 		return;
                 }
 
@@ -546,14 +842,40 @@ void handle_connection(int sock){
 		/* log error to syslog facility */
 		syslog(LOG_ERR,"Data packet from client was too short, bailing out...");
 
+#ifdef HAVE_SSL
+		SSL_shutdown(ssl);
+		SSL_free(ssl);
+#endif
+
 		return;
 	        }
 
-	/* make sure this is the right type of packet */
-	if(ntohl(receive_packet.packet_type)!=QUERY_PACKET || ntohl(receive_packet.packet_version)!=NRPE_PACKET_VERSION_1){
+#ifdef DEBUG
+	fp=fopen("/tmp/packet","w");
+	if(fp){
+		fwrite(&receive_packet,1,sizeof(receive_packet),fp);
+		fclose(fp);
+	        }
+#endif
 
-		/* log error to syslog facility */
-		syslog(LOG_ERR,"Received invalid packet from client, bailing out...");
+	/* make sure the request is valid */
+	if(validate_request(&receive_packet)==ERROR){
+
+		/* log an error */
+		syslog(LOG_ERR,"Client request was invalid, bailing out...");
+
+		/* free memory */
+		free(command_name);
+		command_name=NULL;
+		for(x=0;x<MAX_COMMAND_ARGUMENTS;x++){
+			free(macro_argv[x]);
+			macro_argv[x]=NULL;
+	                }
+
+#ifdef HAVE_SSL
+		SSL_shutdown(ssl);
+		SSL_free(ssl);
+#endif
 
 		return;
 	        }
@@ -563,7 +885,7 @@ void handle_connection(int sock){
 		syslog(LOG_DEBUG,"Host is asking for command '%s' to be run...",receive_packet.buffer);
 
 	/* if this is the version check command, just spew it out */
-	if(!strcmp(&receive_packet.buffer[0],NRPE_HELLO_COMMAND)){
+	if(!strcmp(command_name,NRPE_HELLO_COMMAND)){
 
 		snprintf(buffer,sizeof(buffer),"NRPE v%s",PROGRAM_VERSION);
 		buffer[sizeof(buffer)-1]='\x0';
@@ -577,10 +899,10 @@ void handle_connection(int sock){
 
 	/* find the command we're supposed to run */
 	else{
-		temp_command=find_command(receive_packet.buffer);
+		temp_command=find_command(command_name);
 		if(temp_command==NULL){
 
-			snprintf(buffer,sizeof(buffer),"NRPE: Command '%s' not defined",receive_packet.buffer);
+			snprintf(buffer,sizeof(buffer),"NRPE: Command '%s' not defined",command_name);
 			buffer[sizeof(buffer)-1]='\x0';
 
 			/* log error to syslog facility */
@@ -592,13 +914,22 @@ void handle_connection(int sock){
 
 		else{
 
+			/* process command line */
+			strncpy(raw_command,temp_command->command_line,sizeof(raw_command)-1);
+			raw_command[sizeof(raw_command)-1]='\x0';
+			process_macros(raw_command,processed_command,sizeof(processed_command));
+
 			/* log info to syslog facility */
 			if(debug==TRUE)
-				syslog(LOG_DEBUG,"Running command: %s",temp_command->command_line);
+				syslog(LOG_DEBUG,"Running command: %s",processed_command);
 
 			/* run the command */
 			strcpy(buffer,"");
-			result=my_system(temp_command->command_line,command_timeout,&early_timeout,buffer,sizeof(buffer));
+			result=my_system(processed_command,command_timeout,&early_timeout,buffer,sizeof(buffer));
+
+			/* log debug info */
+			if(debug==TRUE)
+				syslog(LOG_DEBUG,"Command completed with return code %d and output: %s",result,buffer);
 
 			/* see if the command timed out */
 			if(early_timeout==TRUE)
@@ -619,6 +950,14 @@ void handle_connection(int sock){
 		        }
 	        }
 
+	/* free memory */
+	free(command_name);
+	command_name=NULL;
+	for(x=0;x<MAX_COMMAND_ARGUMENTS;x++){
+		free(macro_argv[x]);
+		macro_argv[x]=NULL;
+	        }
+
 	/* strip newline character from end of output buffer */
 	if(buffer[strlen(buffer)-1]=='\n')
 		buffer[strlen(buffer)-1]='\x0';
@@ -626,17 +965,38 @@ void handle_connection(int sock){
 	/* clear the response packet buffer */
 	bzero(&send_packet,sizeof(send_packet));
 
-	/* fill the response packet with data */
-	send_packet.packet_type=htonl(RESPONSE_PACKET);
-	send_packet.packet_version=htonl(NRPE_PACKET_VERSION_1);
-	send_packet.result_code=htonl(result);
-	send_packet.buffer_length=htonl(strlen(buffer));
-	strncpy(&send_packet.buffer[0],buffer,sizeof(send_packet.buffer));
-	send_packet.buffer[sizeof(send_packet.buffer)-1]='\x0';
+	/* fill the packet with semi-random data */
+	randomize_buffer((char *)&send_packet,sizeof(send_packet));
+
+	/* initialize response packet data */
+	send_packet.packet_version=(int16_t)htons(NRPE_PACKET_VERSION_2);
+	send_packet.packet_type=(int16_t)htons(RESPONSE_PACKET);
+	send_packet.result_code=(int16_t)htons(result);
+	strncpy(&send_packet.buffer[0],buffer,MAX_PACKETBUFFER_LENGTH);
+	send_packet.buffer[MAX_PACKETBUFFER_LENGTH-1]='\x0';
 	
+	/* calculate the crc 32 value of the packet */
+	send_packet.crc32_value=(u_int32_t)0L;
+	calculated_crc32=calculate_crc32((char *)&send_packet,sizeof(send_packet));
+	send_packet.crc32_value=(u_int32_t)htonl(calculated_crc32);
+
+
+	/***** ENCRYPT RESPONSE *****/
+
+
 	/* send the response back to the client */
 	bytes_to_send=sizeof(send_packet);
-	sendall(sock,(char *)&send_packet,&bytes_to_send);
+	if(use_ssl==FALSE)
+		sendall(sock,(char *)&send_packet,&bytes_to_send);
+#ifdef HAVE_SSL
+	else
+		SSL_write(ssl,&send_packet,bytes_to_send);
+#endif
+
+#ifdef HAVE_SSL
+	SSL_shutdown(ssl);
+	SSL_free(ssl);
+#endif
 
 	/* log info to syslog facility */
 	if(debug==TRUE)
@@ -647,24 +1007,6 @@ void handle_connection(int sock){
 
 
 
-/* checks to see if a given host is allowed to talk to us */
-int is_an_allowed_host(char *connecting_host){
-	char temp_buffer[MAX_INPUT_BUFFER];
-	char *temp_ptr;
-
-	strncpy(temp_buffer,allowed_hosts,sizeof(temp_buffer));
-	temp_buffer[sizeof(temp_buffer)-1]='\x0';
-
-	for(temp_ptr=strtok(temp_buffer,",");temp_ptr!=NULL;temp_ptr=strtok(NULL,",")){
-		if(!strcmp(connecting_host,temp_ptr))
-			return 1;
-	        }
-
-	return 0;
-        }
-
-
-
 /* handle signals */
 void sighandler(int sig){
 
@@ -689,6 +1031,8 @@ void free_memory(void){
 	this_command=command_list;
 	while(this_command!=NULL){
 		next_command=this_command->next;
+		free(this_command->command_name);
+		free(this_command->command_line);
 		free(this_command);
 		this_command=next_command;
 	        }
@@ -706,6 +1050,7 @@ int my_system(char *command,int timeout,int *early_timeout,char *output,int outp
 	int result;
 	extern int errno;
 	char buffer[MAX_INPUT_BUFFER];
+	char temp_buffer[MAX_INPUT_BUFFER];
 	int fd[2];
 	FILE *fp;
 	int bytes_read=0;
@@ -784,6 +1129,10 @@ int my_system(char *command,int timeout,int *early_timeout,char *output,int outp
 			strcpy(buffer,"");
 			fgets(buffer,sizeof(buffer)-1,fp);
 
+			/* ADDED 01/19/2004 */
+			/* ignore any additional lines of output */
+			while(fgets(temp_buffer,sizeof(temp_buffer)-1,fp));
+
 			/* close the command and get termination status */
 			status=pclose(fp);
 
@@ -831,6 +1180,7 @@ int my_system(char *command,int timeout,int *early_timeout,char *output,int outp
 
 		/* try and read the results from the command output (retry if we encountered a signal) */
 		if(output!=NULL){
+			strcpy(output,"");
 			do{
 				bytes_read=read(fd[0],output,output_length-1);
 		                }while(bytes_read==-1 && errno==EINTR);
@@ -887,6 +1237,7 @@ int drop_privileges(char *user, char *group){
 				gid=(gid_t)(grp->gr_gid);
 			else
 				syslog(LOG_ERR,"Warning: Could not get group entry for '%s'",group);
+			endgrent();
 		        }
 
 		/* else we were passed the GID */
@@ -912,6 +1263,7 @@ int drop_privileges(char *user, char *group){
 				uid=(uid_t)(pw->pw_uid);
 			else
 				syslog(LOG_ERR,"Warning: Could not get passwd entry for '%s'",user);
+			endpwent();
 		        }
 
 		/* else we were passed the UID */
@@ -942,46 +1294,317 @@ int drop_privileges(char *user, char *group){
         }
 
 
+/* write an optional pid file */
+int write_pid_file(void){
+	int fd;
+	int result=0;
+	pid_t pid=0;
+	char pbuf[16];
+
+	/* no pid file was specified */
+	if(pid_file==NULL)
+		return OK;
+
+	/* read existing pid file */
+	if((fd=open(pid_file,O_RDONLY))>=0){
+
+		result=read(fd,pbuf,(sizeof pbuf)-1);
+
+		close(fd);
+
+		if(result>0){
+
+			pbuf[result]='\x0';
+			pid=(pid_t)atoi(pbuf);
+
+			/* if previous process is no longer running running, remove the old pid file */
+			if(pid && (pid==getpid() || kill(pid,0)<0))
+				unlink(pid_file);
+
+			/* previous process is still running */
+			else{
+				syslog(LOG_ERR,"There's already a nrpe server running.");
+				return ERROR;
+			        }
+		        }
+	        } 
+
+	/* write new pid file */
+	if((fd=open(pid_file,O_WRONLY | O_CREAT,0644))>=0){
+		sprintf(pbuf,"%d\n",(int)getpid());
+		write(fd,pbuf,strlen(pbuf));
+		close(fd);
+	        }
+	else{
+		syslog(LOG_ERR,"Cannot write to pidfile '%s'.",pid_file);
+	        }
+
+	return OK;
+        }
+
+
+
+/* bail if daemon is running as root */
+int check_privileges(void){
+	uid_t uid=-1;
+	gid_t gid=-1;
+
+	uid=geteuid();
+	gid=getegid();
+
+	if(uid==0 || gid==0){
+		syslog(LOG_ERR,"Error: NRPE daemon cannot be run as user/group root!");
+		exit(STATE_CRITICAL);
+	        }
+
+	return OK;
+        }
+
+
+
+/* tests whether or not a client request is valid */
+int validate_request(packet *pkt){
+        u_int32_t packet_crc32;
+        u_int32_t calculated_crc32;
+	char *ptr;
+#ifdef ENABLE_COMMAND_ARGUMENTS
+	int x;
+#endif
+
+
+	/***** DECRYPT REQUEST ******/
+
+
+        /* check the crc 32 value */
+        packet_crc32=ntohl(pkt->crc32_value);
+        pkt->crc32_value=0L;
+        calculated_crc32=calculate_crc32((char *)pkt,sizeof(packet));
+        if(packet_crc32!=calculated_crc32){
+                syslog(LOG_ERR,"Error: Request packet had invalid CRC32.");
+                return ERROR;
+                }
+
+	/* make sure this is the right type of packet */
+	if(ntohs(pkt->packet_type)!=QUERY_PACKET || ntohs(pkt->packet_version)!=NRPE_PACKET_VERSION_2){
+		syslog(LOG_ERR,"Error: Request packet type/version was invalid!");
+		return ERROR;
+	        }
+
+	/* make sure buffer is terminated */
+	pkt->buffer[MAX_PACKETBUFFER_LENGTH-1]='\x0';
+
+	/* client must send some kind of request */
+	if(!strcmp(pkt->buffer,"")){
+		syslog(LOG_ERR,"Error: Request contained no query!");
+		return ERROR;
+	        }
+
+	/* make sure request doesn't contain nasties */
+	if(contains_nasty_metachars(pkt->buffer)==TRUE){
+		syslog(LOG_ERR,"Error: Request contained illegal metachars!");
+		return ERROR;
+	        }
+
+	/* make sure the request doesn't contain arguments */
+	if(strchr(pkt->buffer,'!')){
+#ifdef ENABLE_COMMAND_ARGUMENTS
+		if(allow_arguments==FALSE){
+			syslog(LOG_ERR,"Error: Request contained command arguments, but argument option is not enabled!");
+			return ERROR;
+	                }
+#else
+		syslog(LOG_ERR,"Error: Request contained command arguments!");
+		return ERROR;
+#endif
+	        }
+
+	/* get command name */
+#ifdef ENABLE_COMMAND_ARGUMENTS
+	ptr=strtok(pkt->buffer,"!");
+#else
+	ptr=pkt->buffer;
+#endif	
+	command_name=strdup(ptr);
+	if(command_name==NULL){
+		syslog(LOG_ERR,"Error: Memory allocation failed");
+		return ERROR;
+	        }
+
+#ifdef ENABLE_COMMAND_ARGUMENTS
+	/* get command arguments */
+	if(allow_arguments==TRUE){
+
+		for(x=0;x<MAX_COMMAND_ARGUMENTS;x++){
+			ptr=strtok(NULL,"!");
+			if(ptr==NULL)
+				break;
+			macro_argv[x]=strdup(ptr);
+			if(macro_argv[x]==NULL){
+				syslog(LOG_ERR,"Error: Memory allocation failed");
+				return ERROR;
+			        }
+			if(!strcmp(macro_argv[x],"")){
+				syslog(LOG_ERR,"Error: Request contained an empty command argument");
+				return ERROR;
+		                }
+		        }
+	        }
+#endif
+
+	return OK;
+        }
+
+
+
+/* tests whether a buffer contains illegal metachars */
+int contains_nasty_metachars(char *str){
+	int result;
+
+	if(str==NULL)
+		return FALSE;
+	
+	result=strcspn(str,NASTY_METACHARS);
+	if(result!=strlen(str))
+		return TRUE;
+
+	return FALSE;
+        }
+
+
+
+/* replace macros in buffer */
+int process_macros(char *input_buffer,char *output_buffer,int buffer_length){
+	char *temp_buffer;
+	int in_macro;
+	int arg_index=0;
+	char *selected_macro=NULL;
+
+	strcpy(output_buffer,"");
+
+	in_macro=FALSE;
+
+	for(temp_buffer=my_strsep(&input_buffer,"$");temp_buffer!=NULL;temp_buffer=my_strsep(&input_buffer,"$")){
+
+		selected_macro=NULL;
+
+		if(in_macro==FALSE){
+			if(strlen(output_buffer)+strlen(temp_buffer)<buffer_length-1){
+				strncat(output_buffer,temp_buffer,buffer_length-strlen(output_buffer)-1);
+				output_buffer[buffer_length-1]='\x0';
+			        }
+			in_macro=TRUE;
+			}
+		else{
+
+			if(strlen(output_buffer)+strlen(temp_buffer)<buffer_length-1){
+
+				/* argument macro */
+				if(strstr(temp_buffer,"ARG")==temp_buffer){
+					arg_index=atoi(temp_buffer+3);
+					if(arg_index>=1 && arg_index<=MAX_COMMAND_ARGUMENTS)
+						selected_macro=macro_argv[arg_index-1];
+				        }
+
+				/* an escaped $ is done by specifying two $$ next to each other */
+				else if(!strcmp(temp_buffer,"")){
+					strncat(output_buffer,"$",buffer_length-strlen(output_buffer)-1);
+				        }
+				
+				/* a non-macro, just some user-defined string between two $s */
+				else{
+					strncat(output_buffer,"$",buffer_length-strlen(output_buffer)-1);
+					output_buffer[buffer_length-1]='\x0';
+					strncat(output_buffer,temp_buffer,buffer_length-strlen(output_buffer)-1);
+					output_buffer[buffer_length-1]='\x0';
+					strncat(output_buffer,"$",buffer_length-strlen(output_buffer)-1);
+				        }
+
+				
+				/* insert macro */
+				if(selected_macro!=NULL)
+					strncat(output_buffer,(selected_macro==NULL)?"":selected_macro,buffer_length-strlen(output_buffer)-1);
+
+				output_buffer[buffer_length-1]='\x0';
+				}
+
+			in_macro=FALSE;
+			}
+		}
+
+	return OK;
+	}
+
+
+
 /* process command line arguments */
 int process_arguments(int argc, char **argv){
-	int x;
+	char optchars[MAX_INPUT_BUFFER];
+	int c=1;
 	int have_mode=FALSE;
 
+#ifdef HAVE_GETOPT_LONG
+	int option_index=0;
+	static struct option long_options[]={
+		{"config", required_argument, 0, 'c'},
+		{"inetd", no_argument, 0, 'i'},
+		{"daemon", no_argument, 0, 'd'},
+		{"no-ssl", no_argument, 0, 'n'},
+		{"help", no_argument, 0, 'h'},
+		{"license", no_argument, 0, 'l'},
+		{0, 0, 0, 0}
+                };
+#endif
+
 	/* no options were supplied */
 	if(argc<2)
 		return ERROR;
 
-	/* process all arguments */
-	for(x=2;x<=argc;x++){
+	snprintf(optchars,MAX_INPUT_BUFFER,"c:hVldin");
 
-		if(!strcmp(argv[x-1],"-c")){
-			if(x<argc){
-				strncpy(config_file,argv[x],sizeof(config_file)-1);
-				config_file[sizeof(config_file)-1]='\x0';
-				x++;
-			        }
-			else
-				return ERROR;
-		        }
-		else if(!strcmp(argv[x-1],"-d") || !strcmp(argv[x-1],"--daemon")){
+	while(1){
+#ifdef HAVE_GETOPT_LONG
+		c=getopt_long(argc,argv,optchars,long_options,&option_index);
+#else
+		c=getopt(argc,argv,optchars);
+#endif
+		if(c==-1 || c==EOF)
+			break;
+
+		/* process all arguments */
+		switch(c){
+
+		case '?':
+		case 'h':
+			show_help=TRUE;
+			break;
+		case 'V':
+			show_version=TRUE;
+			break;
+		case 'l':
+			show_license=TRUE;
+			break;
+		case 'c':
+			strncpy(config_file,optarg,sizeof(config_file));
+			config_file[sizeof(config_file)-1]='\x0';
+			break;
+		case 'd':
 			use_inetd=FALSE;
 			have_mode=TRUE;
-		        }
-		else if(!strcmp(argv[1],"-i") || !strcmp(argv[x-1],"--inetd")){
+			break;
+		case 'i':
 			use_inetd=TRUE;
 			have_mode=TRUE;
-		        }
-		else if(!strcmp(argv[x-1],"-h") || !strcmp(argv[x-1],"--help"))
-			show_help=TRUE;
-		else if(!strcmp(argv[x-1],"-l") || !strcmp(argv[x-1],"--license"))
-			show_license=TRUE;
-		else if(!strcmp(argv[x-1],"-V") || !strcmp(argv[x-1],"--version"))
-			show_version=TRUE;
-		else
+			break;
+		case 'n':
+			use_ssl=FALSE;
+			break;
+		default:
 			return ERROR;
+			break;
+		        }
 	        }
 
-	/* bail if we didn't get a mode */
+	/* bail if we didn't get required args */
 	if(have_mode==FALSE)
 		return ERROR;
 
